// Matrix3d.cpp: implementation of the Matrix3d class.
/*
 * Copyright (c) 2009-2018 3Dconnexion. All rights reserved.
 *
 * This file and source code are an integral part of the "3Dconnexion
 * Software Developer Kit", including all accompanying documentation,
 * and is protected by intellectual property laws. All use of the
 * 3Dconnexion Software Developer Kit is subject to the License
 * Agreement found in the "LicenseAgreementSDK.txt" file.
 * All rights not expressly granted by 3Dconnexion are reserved.
 */
///////////////////////////////////////////////////////////////////////////////
///
/// \file
/// This file contains the definition of the Matrix3d class.
///
#include "stdafx.h"
///////////////////////////////////////////////////////////////////////////////
// History
//
// $Id: Matrix3d.cpp 15018 2018-05-17 13:02:47Z mbonk $
//
//
#include "Matrix3d.h"

// stdlib
#include <cmath>

// atl
#include <atltrace.h>

Matrix3d::Matrix3d() { Identity(); }

Matrix3d::~Matrix3d() {}

Matrix3d &Matrix3d::Transpose() {
  double tmp;
#define SWAP(a, b)                                                             \
  tmp = a;                                                                     \
  a = b;                                                                       \
  b = tmp;
  SWAP(m[1][0], m[0][1]);
  SWAP(m[2][0], m[0][2]);
  SWAP(m[3][0], m[0][3]);
  SWAP(m[2][1], m[1][2]);
  SWAP(m[3][1], m[1][3]);
  SWAP(m[3][2], m[2][3]);
#undef SWAP
  return *this;
}

Matrix3d &Matrix3d::TransposeRotOnly() {
  double tmp;
#define SWAP(a, b)                                                             \
  tmp = a;                                                                     \
  a = b;                                                                       \
  b = tmp;
  SWAP(m[1][0], m[0][1]);
  SWAP(m[2][0], m[0][2]);
  SWAP(m[2][1], m[1][2]);
#undef SWAP
  return *this;
}

Matrix3d Matrix3d::Inverse() const {
  Matrix3d result = *this;

  // Orientation is just a transpose
  result.TransposeRotOnly();

  // Position is negative of original position dotted with corresponding
  // orientation vector.
  result.m[0][3] =
      (-m[0][3] * m[0][0]) + (-m[1][3] * m[1][0]) + (-m[2][3] * m[2][0]);
  result.m[1][3] =
      (-m[0][3] * m[0][1]) + (-m[1][3] * m[1][1]) + (-m[2][3] * m[2][1]);
  result.m[2][3] =
      (-m[0][3] * m[0][2]) + (-m[1][3] * m[1][2]) + (-m[2][3] * m[2][2]);

  // Zero out last row
  result.m[3][0] = result.m[3][1] = result.m[3][2] = 0.;

  return result;
}

Matrix3d &Matrix3d::Invert() // inverts inplace
{
  *this = this->Inverse();
  return *this;
}

void Matrix3d::GetEulerAngle(double &x, double &y, double &z,
                             rotationOrder_t order) {
  switch (order) {
  case kXYZ:
    if ((1. - fabs(m[2][0])) < kEpsilon5) {
      z = 0.; // Arbitary
      if (m[2][0] < 0) {
        y = acos(0.);
        x = z + atan2(m[0][1], m[0][2]);
      } else {
        y = -acos(0.);
        x = -z + atan2(-m[0][1], -m[0][2]);
      }
    } else {
      y = -asin(m[2][0]);
      x = atan2(m[2][1] / cos(y), m[2][2] / cos(y));
      z = atan2(m[1][0] / cos(y), m[0][0] / cos(y));
    }
    break;

  case kXZY:
    if ((1. - fabs(m[1][0])) < kEpsilon5) {
      y = 0.; // Arbitary
      if (m[1][0] < 0) {
        z = -acos(0.);
        x = y - atan2(m[0][2], m[0][1]);
      } else {
        z = acos(0.);
        x = -y + atan2(m[0][2], -m[0][1]);
      }
    } else {
      z = asin(m[1][0]);
      y = atan2(-m[2][0] / cos(z), m[0][0] / cos(z));
      x = atan2(-m[1][2] / cos(z), m[1][1] / cos(z));
    }
    break;

  case kYXZ:
    if ((1. - fabs(m[2][1])) < kEpsilon5) {
      z = 0.; // Arbitary
      if (m[2][1] < 0) {
        x = -acos(0.);
        y = z - atan2(m[1][0], m[0][0]);
      } else {
        x = acos(0.);
        y = z + atan2(m[1][0], m[0][0]);
      }
    } else {
      x = asin(m[2][1]);
      y = atan2(-m[2][0], m[2][2]);
      z = atan2(-m[0][1], m[1][1]);
    }
    break;
  default:
    x = y = z = 0;
  }
}

double Matrix3d::GetRotationAxisAngle(Vector3d &axis) {
  double cosAngle = (m[0][0] + m[1][1] + m[2][2] - 1.) / 2.;
  double angle =
      cosAngle > 1. ? acos(1.) : cosAngle < -1. ? acos(-1.) : acos(cosAngle);

  const double epsilon2 = 0.1;
  if (fabs(angle) < kEpsilon5 || fabs(angle - acos(-1.)) < kEpsilon5) {
    if (fabs(m[0][1] + m[1][0]) < epsilon2 &&
        fabs(m[0][2] + m[2][0]) < epsilon2 &&
        fabs(m[1][2] + m[2][1]) < epsilon2 &&
        fabs(m[0][0] + m[1][1] + m[2][2] - 3) < epsilon2) {
      // this singularity is identity matrix so angle = 0
      axis = Vector3d(1., 0., 0.);
      return 0.; // zero angle, arbitrary axis
    }
    angle = acos(-1.);
    double xx = (m[0][0] + 1) / 2;
    double yy = (m[1][1] + 1) / 2;
    double zz = (m[2][2] + 1) / 2;
    double xy = (m[0][1] + m[1][0]) / 4;
    double xz = (m[0][2] + m[2][0]) / 4;
    double yz = (m[1][2] + m[2][1]) / 4;
    if ((xx > yy) && (xx > zz)) { // m[0][0] is the largest diagonal term
      if (xx < kEpsilon5)
        axis = Vector3d(0., sqrt(1. / 2.), sqrt(1. / 2.));
      else
        axis = Vector3d(sqrt(xx), xy / sqrt(xx), xz / sqrt(xx));
    } else if (yy > zz) { // m[1][1] is the largest diagonal term
      if (yy < kEpsilon5)
        axis = Vector3d(sqrt(1. / 2.), 0., sqrt(1. / 2.));
      else
        axis = Vector3d(xy / sqrt(yy), sqrt(yy), yz / sqrt(yy));
    } else { // m[2][2] is the largest diagonal term so base result on this
      if (zz < kEpsilon5)
        axis = Vector3d(sqrt(1. / 2.), sqrt(1. / 2.), 0.);
      else
        axis = Vector3d(xz / sqrt(zz), yz / sqrt(zz), sqrt(zz));
    }
    return angle;
  }
  // as we have reached here there are no singularities so we can handle
  // normally
  double s =
      sqrt((m[2][1] - m[1][2]) * (m[2][1] - m[1][2]) +
           (m[0][2] - m[2][0]) * (m[0][2] - m[2][0]) +
           (m[1][0] - m[0][1]) * (m[1][0] - m[0][1])); // used to normalise
  if (fabs(s) < kEpsilon5)
    s = 1.;
  axis = Vector3d((m[2][1] - m[1][2]) / s, (m[0][2] - m[2][0]) / s,
                  (m[1][0] - m[0][1]) / s);
  return angle;
}

Matrix3d &Matrix3d::SetFromAngleAxis(const Vector3d &axis) {
  return SetFromAngleAxis(axis.Length(), axis.UnitVector());
}

Matrix3d &Matrix3d::SetFromAngleAxis(double angle, const Vector3d &axis) {
  if (fabs(angle) < kEpsilon5) {
    Identity();
    return *this;
  }

  double x = axis.v[0], y = axis.v[1], z = axis.v[2], c = cos(angle),
         s = sin(angle), v = 1. - c;

  m[0][0] = x * x * v + c;
  m[0][1] = x * y * v - z * s;
  m[0][2] = x * z * v + y * s;
  m[0][3] = 0.;

  m[1][0] = y * x * v + z * s;
  m[1][1] = y * y * v + c;
  m[1][2] = y * z * v - x * s;
  m[1][3] = 0.;

  m[2][0] = z * x * v - y * s;
  m[2][1] = z * y * v + x * s;
  m[2][2] = z * z * v + c;
  m[2][3] = 0.;

  m[3][0] = m[3][1] = m[3][2] = 0.;
  m[3][3] = 1.;

  return *this;
}

void Matrix3d::DumpMatrix(char *s) {
  _RPT1(_CRT_WARN, "\n%s\n", s);
  _RPT4(_CRT_WARN, "%f %f %f %f\n", m[0][0], m[0][1], m[0][2], m[0][3]);
  _RPT4(_CRT_WARN, "%f %f %f %f\n", m[1][0], m[1][1], m[1][2], m[1][3]);
  _RPT4(_CRT_WARN, "%f %f %f %f\n", m[2][0], m[2][1], m[2][2], m[2][3]);
  _RPT4(_CRT_WARN, "%f %f %f %f\n", m[3][0], m[3][1], m[3][2], m[3][3]);
}

Matrix3d Matrix3d::operator*(const Matrix3d &mat) const {
  Matrix3d result = *this;
  result *= mat;
  return result;
}

Matrix3d &Matrix3d::operator*=(const Matrix3d &mat) {
  Matrix3d tmp;
  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
      tmp.m[i][j] = m[i][0] * mat.m[0][j] + m[i][1] * mat.m[1][j] +
                    m[i][2] * mat.m[2][j] + m[i][3] * mat.m[3][j];

  *this = tmp;
  return *this;
}

Matrix3d Matrix3d::operator+(const Matrix3d &mat) const {
  Matrix3d result = *this;
  result += mat;
  return result;
}

Matrix3d &Matrix3d::operator+=(const Matrix3d &mat) {
  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
      m[i][j] += mat.m[i][j];
  return *this;
}

Matrix3d Matrix3d::operator-(const Matrix3d &mat) const {
  Matrix3d result = *this;
  result -= mat;
  return result;
}

Matrix3d &Matrix3d::operator-=(const Matrix3d &mat) {
  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
      m[i][j] -= mat.m[i][j];
  return *this;
}

Vector3d Matrix3d::operator*(const Vector3d &vec) const {
  Vector3d result;
  result.v[0] = m[0][0] * vec.v[0] + m[0][1] * vec.v[1] + m[0][2] * vec.v[2];
  result.v[1] = m[1][0] * vec.v[0] + m[1][1] * vec.v[1] + m[1][2] * vec.v[2];
  result.v[2] = m[2][0] * vec.v[0] + m[2][1] * vec.v[1] + m[2][2] * vec.v[2];
  result.v[3] = m[3][0] * vec.v[0] + m[3][1] * vec.v[1] + m[3][2] * vec.v[2] +
                m[3][3] * vec.v[3];
  return result;
}

Point3d Matrix3d::operator*(const Point3d &pos) const {
  Point3d result;
  result.p[0] = m[0][0] * pos.p[0] + m[0][1] * pos.p[1] + m[0][2] * pos.p[2] +
                m[0][3] * pos.p[3];
  result.p[1] = m[1][0] * pos.p[0] + m[1][1] * pos.p[1] + m[1][2] * pos.p[2] +
                m[1][3] * pos.p[3];
  result.p[2] = m[2][0] * pos.p[0] + m[2][1] * pos.p[1] + m[2][2] * pos.p[2] +
                m[2][3] * pos.p[3];
  result.p[3] = m[3][0] * pos.p[0] + m[3][1] * pos.p[1] + m[3][2] * pos.p[2] +
                m[3][3] * pos.p[3];
  return result;
}

bool Matrix3d::operator==(const Matrix3d &mat) const {
  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
      if (fabs(m[i][j] - mat.m[i][j]) > kEpsilon5)
        return false;
  return true;
}

Matrix3d::operator matrix16_t const &() const {
  return *reinterpret_cast<const matrix16_t *>(m);
}

Matrix3d::operator matrix4x4_t const &() const { return m; }

void Matrix3d::Test() {
  // Test vector addition
  Vector3d v1(1, 0, 0), v2(0, 1, 0);
  Vector3d v3 = v1 + v2;
  v3.Dump("v3=v1+v2");
  v3 += v1;
  v3.Dump("v3+=v1");

  // Test vector subtraction
  v3 -= v1;
  v3.Dump("v3-=v1");
  v3 = v1 - v2;
  v3.Dump("v3=v1-v2");

  // Test vector cross product
  Vector3d cross = v1.CrossProduct(v2);
  cross.Dump("v1 x v2");

  // Test vector dot product
  _RPT1(_CRT_WARN, "v1 dot v2: %f\n", v1.DotProduct(v2));

  Vector3d v4(.5, .5, .5);
  _RPT1(_CRT_WARN, "v1 dot v4: %f\n", v1.DotProduct(v4));

  // Test vector zero
  v4.Zero();
  v4.Dump("zero");

  // Test vector set
  v4.Set(1, 2, 3);
  v4.Dump("v4.Set(1,2,3)");

  // Test matrix transpose
  Matrix3d t1(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16);
  t1.DumpMatrix("original t1");
  t1.Transpose();
  t1.DumpMatrix("transposed t1");

  // Test matrix addition
  Matrix3d t2(2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30, 32);
  Matrix3d t3;
  t3 = t1 + t2;
  t3.DumpMatrix("t3 = t1+t2");

  // Test matrix subtraction
  t1 = t3 - t2;
  t1.DumpMatrix("t1 = t3-t2");

  // Test matrix multiplication
  Matrix3d I;
  I.Identity();
  t1 *= I;
  t1.DumpMatrix("t1 * I");
  Matrix3d t4 = t1;
  t4 = t4 * t1;
  t4.DumpMatrix("t4=t4*t1");
  t4 = t1;
  t4 *= t1;
  t4.DumpMatrix("t4*=t1");

  // Test matrix inversion
  t1.Set(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16);
  Matrix3d t1Invert = t1;
  t1Invert.Invert();
  t1Invert.DumpMatrix("t1Invert:");

  Matrix3d t1Inv = t1.Inverse();
  t1Inv.DumpMatrix("t1Inv");

  // does a * INV(a) = I?
  Vector3d rot(10, 0, 0);
  t1.SetFromAngleAxis(rot);
  t1Inv = t1.Inverse();
  t1.DumpMatrix("t1 (a rotation)");
  t1Inv.DumpMatrix("t1Inverse");
  t1 *= t1Inv;
  t1.DumpMatrix("t1 *= t1Inv (should be I)");
}

Matrix3d &Matrix3d::TranslateBy(const Vector3d &v) {
  m[0][3] += v.v[0];
  m[1][3] += v.v[1];
  m[2][3] += v.v[2];
  return *this;
}

Matrix3d &Matrix3d::SetPosition(const Point3d &pos) {
  m[0][3] = pos.p[0] * pos.p[3] / m[3][3];
  m[1][3] = pos.p[1] * pos.p[3] / m[3][3];
  m[2][3] = pos.p[2] * pos.p[3] / m[3][3];
  return *this;
}

Matrix3d &Matrix3d::SetOrientation(const Matrix3d &mat) {
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      m[i][j] = mat.m[i][j];
  return *this;
}

Vector3d Matrix3d::GetColumn(size_t i) {
  Vector3d column;
  if (i < 4) {
    column.v[0] = m[0][i];
    column.v[1] = m[1][i];
    column.v[2] = m[2][i];
    column.v[3] = m[3][i];
  }
  return column;
}

Point3d Matrix3d::GetPosition() {
  return Point3d(m[0][3], m[1][3], m[2][3], m[3][3]);
}

Matrix3d Matrix3d::GetOrientation() {
  Matrix3d orientation;
  for (int i = 0; i < 3; i++)
    for (int j = 0; j < 3; j++)
      orientation.m[i][j] = m[i][j];
  return orientation;
}

Matrix3d &Matrix3d::SetFromScaleFactors(double sx, double sy, double sz) {
  m[0][0] = sx;
  m[0][1] = 0;
  m[0][2] = 0;
  m[0][3] = 0;
  m[1][0] = 0;
  m[1][1] = sy;
  m[1][2] = 0;
  m[1][3] = 0;
  m[2][0] = 0;
  m[2][1] = 0;
  m[2][2] = sz;
  m[2][3] = 0;
  m[3][0] = 0;
  m[3][1] = 0;
  m[3][2] = 0;
  m[3][3] = 1;
  return *this;
}